@@ -7,7 +7,7 @@
    Manger and several other people whose names I've misplaced.
 
    Available from http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.c. Last
-   updated 12 November 2010 (version 20101112, if you prefer it that way).  
+   updated 21 May 2012 (version 20120521, if you prefer it that way).  
    To build under Windows, use 'cl /MD dumpasn1.c'.  To build on OS390 or 
    z/OS, use '/bin/c89 -D OS390 -o dumpasn1 dumpasn1.c'.
 
@@ -56,7 +56,7 @@
 
 /* The update string, printed as part of the help screen */
 
-#define UPDATE_STRING	"12 November 2010"
+#define UPDATE_STRING	"21 May 2012"
 
 /* Useful defines */
 
@@ -105,6 +105,10 @@
 #if ( defined( _WINDOWS ) || defined( WIN32 ) || defined( _WIN32 ) || \
 	  defined( __WIN32__ ) )
   #include <windows.h>
+  #include <fcntl.h>				/* For _setmode() codes */
+  #ifndef _O_U16TEXT 
+	#define _O_U16TEXT		0x20000	/* _setmode() code */
+  #endif /* !_O_U16TEXT */
   #define __WIN32__
 #endif /* Win32 */
 
@@ -245,7 +249,7 @@ typedef struct {
 	long length;				/* Data length */
 	int indefinite;				/* Item has indefinite length */
 	int headerSize;				/* Size of tag+length */
-	unsigned char header[ 8 ];	/* Tag+length data */
+	unsigned char header[ 16 ];	/* Tag+length data */
 	} ASN1_ITEM;
 
 /* Configuration options */
@@ -272,10 +276,51 @@ static int outputWidth = 80;		/* 80-column display */
 static int maxNestLevel = 100;		/* Maximum nesting level for which to display output */
 static int doOutlineOnly = FALSE;	/* Only display constructed-object outline */
 
-/* The indent size and fixed indent string to the left of the data */
+/* Formatting information used for the fixed informational column to the 
+   left of the displayed data */
 
-#define INDENT_SIZE		11
-#define INDENT_STRING	"         : "
+static int infoWidth = 4;
+static const char *indentStringTbl[] = {
+	NULL, NULL, NULL, 
+	"       : ",			/* "xxx xxx: " (3) */
+	"         : ",			/* "xxxx xxxx: " (4) */
+	"           : ",		/* "xxxxx xxxxx: " (5) */
+	"             : ",		/* "xxxxxx xxxxxx: " (6) */
+	"               : ",	/* "xxxxxxx xxxxxxx: " (7) */
+	"                 : ",	/* "xxxxxxxx xxxxxxxx: " (8) */
+	"", "", "", ""
+	};
+static const char *lenTbl[] = {
+	NULL, NULL, NULL, 
+	"%3ld %3ld: ", "%4ld %4ld: ", "%5ld %5ld: ",
+	"%6ld %6ld: ", "%7ld %7ld: ", "%8ld %8ld: ",
+	"", "", "", ""
+	};
+static const char *lenIndefTbl[] = {
+	NULL, NULL, NULL, 
+	"%3ld NDF: ", "%4ld NDEF: ", "%5ld INDEF: ",
+	"%6ld INDEF : ", "%7ld INDEF  : ", "%8ld INDEF   : ",
+	"", "", "", ""
+	};
+static const char *lenHexTbl[] = {
+	NULL, NULL, NULL, 
+	"%03lX %3lX: ", "%04lX %4lX: ", "%05lX %5lX: ",
+	"%06lX %6lX: ", "%07lX %7lX: ", "%08lX %8lX: ",
+	"", "", "", ""
+	};
+static const char *lenHexIndefTbl[] = {
+	NULL, NULL, NULL, 
+	"%03lX NDF: ", "%04lX NDEF: ", "%05lX INDEF: ",
+	"%06lX INDEF : ", "%07lX INDEF  : ", "%08lX INDEF   : ",
+	"", "", "", ""
+	};
+
+#define INDENT_SIZE		( infoWidth + 1 + infoWidth + 1 + 1 )
+#define INDENT_STRING	indentStringTbl[ infoWidth ]
+#define LEN				lenTbl[ infoWidth ]
+#define LEN_INDEF		lenIndefTbl[ infoWidth ]
+#define LEN_HEX			lenHexTbl[ infoWidth ]
+#define LEN_HEX_INDEF	lenHexIndefTbl[ infoWidth ]
 
 /* Error and warning information */
 
@@ -1160,6 +1205,84 @@ static int adjustLevel( const int level, const int maxLevel )
 	return( level );
 	}
 
+#if defined( __WIN32__ ) || defined( __UNIX__ ) || defined( __OS390__ )
+
+/* Try and display to display a Unicode character.  This is pretty hit and 
+   miss, and if it fails nothing is displayed.  To try and detect this we 
+   use wcstombs() to see if anything can be displayed, if it can't we drop 
+   back to trying to display the data as non-Unicode */
+
+static int displayUnicode( const wchar_t wCh, const int level )
+	{
+	char outBuf[ 8 ];
+	int outLen;
+
+	/* Check whether we can display this character */
+	outLen = wcstombs( outBuf, &wCh, 1 );
+	if( outLen < 1 )
+		{
+		/* Tell the caller that this can't be displayed as Unicode */
+		return( FALSE );
+		}
+
+#if defined( __WIN32__ )
+	if( level < maxNestLevel )
+		{
+		int oldmode;
+						
+		/* To output Unicode to the Win32 console we need to switch the 
+		   output stream to Unicode-16 mode, but the following may also 
+		   depend on which code page is currently set for the console and 
+		   which font is being used */
+		fflush( output );
+		oldmode = _setmode( fileno( output ), _O_U16TEXT );
+		fputwc( wCh, output );
+		_setmode( fileno( output ), oldmode );
+		}
+#elif defined( __UNIX__ ) && !( defined( __MACH__ ) || defined( __OpenBSD__ ) )
+	/* Unix environments are even more broken than Win32, like Win32 the 
+	   output differentiates between char and widechar output, but there's 
+	   no easy way to deal with this.  In theory fwide() can set it, but 
+	   it's a one-way function, once we've set it a particular way we can't 
+	   go back (exactly what level of braindamage it takes to have an 
+	   implementation function like this is a mystery).  Other sources 
+	   suggest using setlocale() tricks, printf() with "%lc" as the format 
+	   specifier, and others, but this doesn't seem to work either */
+	if( level < maxNestLevel )
+		{
+#if 0
+		setlocale( LC_ALL, "" );
+		fputwc( wCh, output );
+#elif 0
+		fprintf( output, "%lc", wCh );
+#else
+		if( fwide( output, 1 ) > 0 )
+			{
+			fputwc( wCh, output );
+			fwide( output, -1 );
+			}
+		else
+			fputc( wCh, output );
+#endif
+		}
+#else
+  #ifdef __OS390__
+	if( level < maxNestLevel )
+		{
+		char *p;
+
+		/* This could use some improvement */
+		for( p = outBuf; *p != '\0'; p++ )
+			*p = asciiToEbcdic( *p );
+		}
+  #endif /* IBM ASCII -> EBCDIC conversion */
+	printString( level, "%s", outBuf );
+#endif /* OS-specific charset handling */
+
+	return( TRUE );
+	}
+#endif /* __WIN32__ || __UNIX__ || __OS390__ */
+
 /* Display an integer value */
 
 static void printValue( FILE *inFile, const int valueLength, const int level )
@@ -1190,7 +1313,7 @@ static void printValue( FILE *inFile, const int valueLength, const int level )
 	printString( level, " %ld\n", value );
 	if( warnNonDER )
 		complain( "Integer has non-DER encoding", level );
-	if( warnNegative < 0 )
+	if( warnNegative )
 		complain( "Integer has a negative value", level );
 	}
 
@@ -1199,21 +1322,34 @@ static void printValue( FILE *inFile, const int valueLength, const int level )
 static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 	{
 	const int lineLength = ( dumpText ) ? 8 : 16;
+	const int displayHeaderLength = ( ( doPure ) ? 0 : INDENT_SIZE ) + 2;
 	char printable[ 9 ];
 	long noBytes = length;
 	int warnPadding = FALSE, warnNegative = isInteger, singleLine = FALSE;
-	int prevCh = -1, i;
+	int displayLength = displayHeaderLength, prevCh = -1, i;
 
-	/* Check if LHS status info + indent + "OCTET STRING" string + data will
-	   wrap */
-	if( ( ( doPure ) ? 0 : INDENT_SIZE ) + ( level * 2 ) + 12 + \
-		( length * 3 ) < outputWidth )
+	memset( printable, 0, 9 );
+
+	displayLength += ( length < lineLength ) ? ( length * 3 ) : \
+											   ( lineLength * 3 );
+
+	/* Check if the size of the displayed data (LHS status info + hex data) 
+	   plus the indent-level of spaces will fit into a single line behind 
+	   the initial label, e.g. "INTEGER" */
+	if( displayHeaderLength + ( level * 2 ) + ( length * 3 ) < outputWidth )
 		singleLine = TRUE;
 
-	if( noBytes > 128 && !printAllData )
-		noBytes = 128;	/* Only output a maximum of 128 bytes */
-	level = adjustLevel( level, ( doPure ) ? 15 : 8 );
-	printable[ 8 ] = printable[ 0 ] = '\0';
+	/* By default we only output a maximum of 128 bytes to avoid dumping 
+	   huge amounts of data, however if what's left is a partial lines' 
+	   worth then we output that as well to avoid displaying a line of text 
+	   indicating that less than a lines' worth of data remains to be 
+	   displayed */
+	if( noBytes >= 128 + lineLength && !printAllData )
+		noBytes = 128;
+
+	/* Make sure that the indent level doesn't push the text off the edge of 
+	   the screen */
+	level = adjustLevel( level, ( outputWidth - displayLength ) / 2 );
 	for( i = 0; i < noBytes; i++ )
 		{
 		int ch;
@@ -1238,7 +1374,7 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 				}
 			}
 		ch = getc( inFile );
-		printString( level, "%s%02X", i % lineLength ? " " : "", ch );
+		printString( level, "%s%02X", ( i % lineLength ) ? " " : "", ch );
 		printable[ i % 8 ] = ( ch >= ' ' && ch < 127 ) ? ch : '.';
 		fPos++;
 
@@ -1271,7 +1407,7 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 		printString( level, "%s", "    " );
 		printString( level, "%s", printable );
 		}
-	if( length > 128 && !printAllData )
+	if( length >= 128 + lineLength && !printAllData )
 		{
 		length -= 128;
 		printString( level, "%c", '\n' );
@@ -1575,73 +1711,25 @@ static void displayString( FILE *inFile, long length, int level,
 		if( strOption == STR_BMP )
 			{
 			if( i == noBytes - 1 && ( noBytes & 1 ) )
+				{
 				/* Odd-length BMP string, complain */
 				warnBMP = TRUE;
+				}
 			else
 				{
 				const wchar_t wCh = ( ch << 8 ) | getc( inFile );
-				char outBuf[ 8 ];
-#ifdef __OS390__
-				char *p;
-#endif /* OS-specific charset handling */
-				int outLen;
-
-				/* Attempting to display Unicode characters is pretty hit and
-				   miss, and if it fails nothing is displayed.  To try and
-				   detect this we use wcstombs() to see if anything can be
-				   displayed, if it can't we drop back to trying to display
-				   the data as non-Unicode.  There's one exception to this
-				   case, which is for a wrong-endianness Unicode string, for
-				   which the first character looks like a single ASCII char */
-				outLen = wcstombs( outBuf, &wCh, 1 );
-				if( outLen < 1 )
-					{
-					/* Can't be displayed as Unicode, fall back to
-					   displaying it as normal text */
-					ungetc( wCh & 0xFF, inFile );
-					}
-				else
+				
+				if( displayUnicode( wCh, level ) )
 					{
 					lineLength++;
 					i++;	/* We've read two characters for a wchar_t */
-#if defined( __WIN32__ )
-					if( level < maxNestLevel )
-						fputwc( wCh, output );
-#elif defined( __UNIX__ ) && !( defined( __MACH__ ) || defined( __OpenBSD__ ) )
-					/* Some Unix environments differentiate between char
-					   and wide-oriented stdout (!!!), so it's necessary to
-					   manually switch the orientation of stdout to make it
-					   wide-oriented before calling a widechar output
-					   function or nothing will be output (exactly what
-					   level of braindamage it takes to have an
-					   implementation function like this is a mystery).  In
-					   order to safely display widechars, we therefore have
-					   to use the fwide() kludge function to change stdout
-					   modes around the display of the widechar */
-					if( level < maxNestLevel )
-						{
-						if( fwide( output, 1 ) > 0 )
-							{
-							fputwc( wCh, output );
-							fwide( output, -1 );
-							}
-						else
-							fputc( wCh, output );
-						}
-#else
-  #ifdef __OS390__
-					if( level < maxNestLevel )
-						{
-						/* This could use some improvement */
-						for( p = outBuf; *p != '\0'; p++ )
-							*p = asciiToEbcdic( *p );
-						}
-  #endif /* IBM ASCII -> EBCDIC conversion */
-					printString( level, "%s", outBuf );
-#endif /* OS-specific charset handling */
 					fPos += 2;
 					continue;
 					}
+
+				/* The value can't be displayed as Unicode, fall back to 
+				   displaying it as normal text */
+				ungetc( wCh & 0xFF, inFile );
 				}
 			}
 #endif /* __WIN32__ || __UNIX__ || __OS390__ */
@@ -2496,8 +2584,29 @@ static int printAsn1( FILE *inFile, const int level, long length,
 			/* If the length isn't known and the item has a definite length,
 			   set the length to the item's length */
 			if( !item.indefinite )
+				{
 				length = item.headerSize + item.length;
 
+				/* We can also adjust the width of the informational data 
+				   column to maximise the amount of screen real estate (for
+				   lengths less than the default of four) or get rid of 
+				   oversized columns (for lengths greater than four) */
+				if( length < 1000 )
+					infoWidth = 3;
+				else
+				if( length > 9999999 )
+					infoWidth = 8;
+				else
+				if( length > 999999 )
+					infoWidth = 7;
+				else
+				if( length > 99999 )
+					infoWidth = 6;
+				else
+				if( length > 9999 )
+					infoWidth = 5;
+				}
+
 			/* If the input isn't seekable, turn off some options that
 			   require the use of fseek().  This check isn't perfect (some
 			   streams are slightly seekable due to buffering) but it's
@@ -2534,13 +2643,13 @@ static int printAsn1( FILE *inFile, const int level, long length,
 		if( !doPure && !nonOutlineObject )
 			{
 			if( item.indefinite )
-				printString( level, ( doHexValues ) ? "%04lX NDEF: " :
-							 "%4ld NDEF: ", lastPos );
+				printString( level, ( doHexValues ) ? \
+								LEN_HEX_INDEF : LEN_INDEF, lastPos );
 			else
 				{
 				if( !seenEOC )
-					printString( level, ( doHexValues ) ? "%04lX %4lX: " :
-								 "%4ld %4ld: ", lastPos, item.length );
+					printString( level, ( doHexValues ) ? \
+									LEN_HEX : LEN, lastPos, item.length );
 				}
 			}
 
@@ -2625,7 +2734,7 @@ static int printAsn1( FILE *inFile, const int level, long length,
 static void usageExit( void )
 	{
 	puts( "DumpASN1 - ASN.1 object dump/syntax check program." );
-	puts( "Copyright Peter Gutmann 1997 - 2010.  Last updated " UPDATE_STRING "." );
+	puts( "Copyright Peter Gutmann 1997 - 2012.  Last updated " UPDATE_STRING "." );
 	puts( "" );
 
 	puts( "Usage: dumpasn1 [-acdefghilmoprstuvwxz] <file>" );
